/*
 * Main program to test C++ grammar.
 *
 *	17-Jul-98
 *	Manuel Kessler (mlkessle@cip.physik.uni-wuerzburg.de)
 *		Added code to accept file name as commmand line parameter.
 *		Added symmetry functions and error summary.
 */



#include <iostream>
#include "tokens.h"
#include "CParser.h"
typedef ANTLRCommonToken ANTLRToken;
#include "DLGLexer.h"
#include "PBlackBox.h"

class MyParser : public CParser {
public:
	MyParser(ANTLRTokenBuffer *input) : CParser(input) {
    traceOptionValueDefault=0;  // sets reset value
    traceOption(-1);            // sets current value
    }

//	MyParser(ANTLRTokenBuffer *input, TokenType eof) : CParser(input,eof) {;}

	virtual void enterNewLocalScope()
		{ fprintf(stderr, "enterNewLocalScope();\n"); CParser::enterNewLocalScope(); }
	virtual void exitLocalScope()
		{fprintf(stderr, "exitLocalScope();\n"); CParser::exitLocalScope();}

	// Declaration stuff
	virtual void beginDeclaration()
		{fprintf(stderr, "beginDeclaration();\n"); CParser::beginDeclaration();}
	virtual void beginFunctionDefinition()
		{
			fprintf(stderr, "beginFunctionDefinition();\n");
			CParser::beginFunctionDefinition();
		}
	virtual void beginParameterDeclaration()
		{
			fprintf(stderr, "beginParameterDeclaration();\n");
			CParser::beginParameterDeclaration();
		}
	// MK: added for symmetry
	virtual void endParameterDeclaration()
		{
			fprintf(stderr, "endParameterDeclaration();\n");
			CParser::endParameterDeclaration();
		}
	virtual void beginFieldDeclaration()
		{
			fprintf(stderr, "beginFieldDeclaration();\n");
			CParser::beginFieldDeclaration();
		}
	// MK: added for symmetry
	virtual void endFieldDeclaration()
		{
			fprintf(stderr, "endFieldDeclaration();\n");
			CParser::endFieldDeclaration();
		}
	virtual void declarationSpecifier(StorageClass sc,
									  TypeQualifier tq,
									  TypeSpecifier ts)
		{
			fprintf(stderr, "declarationSpecifier();\n");
			CParser::declarationSpecifier(sc,tq,ts);
		}

	// Declarator stuff
	// MK: cv-qualified pointers
	virtual void declaratorPointerTo(TypeQualifier tq)
		{
			fprintf(stderr, "declaratorPointerTo();\n");
			CParser::declaratorPointerTo(tq);
		}
	virtual void declaratorID(char *id)
		{
			fprintf(stderr, "declaratorID();\n");
			CParser::declaratorID(id);
		}
	virtual void declaratorArray()
		{
			fprintf(stderr, "declaratorArray();\n");
			CParser::declaratorArray();
		}
	virtual void declaratorParameterList()
		{
			fprintf(stderr, "declaratorParameterList();\n");
			CParser::declaratorParameterList();
		}
	virtual void declaratorEndParameterList()
		{
			fprintf(stderr, "declaratorEndParameterList();\n");
			CParser::declaratorEndParameterList();
		}
};

int main(int argc, char *argv[])
{
	// MK: modified to accept file name as command line parameter
    FILE * input=stdin;
    const char * filename = NULL;
	if (argc>1 && argv[1][0]!='-') filename = argv[1];
	if (argc>2 && argv[2][0]!='-') filename = argv[2];
    if (filename == NULL || ! (input = fopen(filename, "r")) ) {
       	fprintf(stderr, "Can't open input file");
        return 0;
    } 
	ParserBlackBox<DLGLexer, MyParser, ANTLRToken> p(input);
	if ( (argc>1 && 0 == strcmp("-trace", argv[1])) ||
		 (argc>2 && 0 == strcmp("-trace", argv[2])) ) {
		p.parser()->traceOption(+2);
	}
    p.parser()->translation_unit();
	// MK: print out summary of errors
	if (p.parser()->syntaxErrCount>0) {
		fprintf(stderr, "%d Syntax errors in %s\n",
			p.parser()->syntaxErrCount, input==stdin?"stdin":filename);
	}
    if (input!=stdin) fclose(input);
	return 0;
}
