#include <iostream>
#include "tokens.h"
#include "CParser.h"
typedef ANTLRCommonToken ANTLRToken;
#include "CSymbol.h"

int CParser::
isTypeName(char *s)
{
	CSymbol *cs = (CSymbol *) symbols->lookup(s);
	if ( cs==NULL || cs->getType()!=CSymbol::otType ) return 0;
	return 1;
}

void CParser::
beginDeclaration()
{
}

void CParser::
endDeclaration()
{
}

void CParser::
beginFunctionDefinition()
{
	functionDefinition = 1;
}

void CParser::
endFunctionDefinition()
{
	fprintf(stderr, "parameter scope:\n");
	symbols->dumpScope(stderr);
	symbols->removeScope();
	symbols->restoreScope();
	functionDefinition = 0;
}

void CParser::
beginParameterDeclaration()
{
}

// MK: added for symmetry
void CParser::
endParameterDeclaration()
{
}

void CParser::
beginFieldDeclaration()
{
}

// MK: added for symmetry
void CParser::
endFieldDeclaration()
{
}

void CParser::
declarationSpecifier(StorageClass sc,
					 TypeQualifier tq,
					 TypeSpecifier ts)
{
	_sc = sc;
	_tq = tq;
	_ts = ts;
}

// MK: cv-qualified pointers
void CParser::
declaratorPointerTo(TypeQualifier tq)
{
	_tq=tq;
}

/* Symbols from declarators are added to the symbol table here.  Note
 * that we only care about types for parsing (with semantic predicates)
 * and we don't track variable vs function.  A real C front-end would
 * have to, of course.  The symbol is added to whatever the current
 * scope is in the symbol table.
 */
void CParser::
declaratorID(char *id)
{
	CSymbol *c;
	if ( _sc==scTYPEDEF ) c = new CSymbol(id, CSymbol::otType);
	else c = new CSymbol(id);
	if ( c==NULL ) panic("can't alloc CSymbol");
	symbols->define(id, c);
}

void CParser::
declaratorArray()
{
}

void CParser::
declaratorParameterList()
{
	symbols->saveScope();
}

void CParser::
declaratorEndParameterList()
{
	if ( !functionDefinition ) {
		fprintf(stderr, "parameter scope:\n");
		symbols->dumpScope(stderr);
		symbols->removeScope();
		symbols->restoreScope();
	}
}

void CParser::
enterNewLocalScope()
{
	symbols->saveScope();
}

void CParser::
exitLocalScope()
{
	fprintf(stderr, "local scope:\n");
	symbols->dumpScope(stderr);
	symbols->removeScope();
	symbols->restoreScope();
}

void CParser::
enterExternalScope()
{
}

void CParser::
exitExternalScope()
{
	fprintf(stderr, "global scope:\n");
	symbols->dumpScope(stderr);
	symbols->removeScope();
}

void CParser::
beginAggrDefinition(TypeSpecifier ts, char *tag)
{
	CSymbol *c;

	if ( ts==tsSTRUCT ) c = new CSymbol(tag, CSymbol::otStruct);
	else c = new CSymbol(tag, CSymbol::otUnion);
	if ( c==NULL ) panic("can't alloc CSymbol");
	symbols->define(tag, c);

	symbols->saveScope();	// use the scope to collect list of fields
}

void CParser::
endAggrDefinition()
{
	fprintf(stderr, "list of fields:\n");
	symbols->dumpScope(stderr);
	symbols->removeScope();
	symbols->restoreScope();
}

void CParser::
panic(char *err)
{
	fprintf(stderr, "CParser: %s\n", err);
	exit(-1);
}

// MK: added for error reporting from member functions, don't know how to
// do it better

void CParser::
synError(char * msg, _ANTLRTokenPtr tok)
{
	fprintf(stderr, "line %d: %s before '%s'\n",
		tok->getLine(), msg, tok->getText());
	++syntaxErrCount;
}

/* MK: Check if type specifier ts and type specifier add together are still
 a valid type specifier. I hope I got all cases right, it's somewhat
 confusing, and I only have 'The C programming language', 2nd ed.
 but it should reflect ANSI
 yes, I know, goto is ugly and to be avoided wherever possible, but
 here I feel its use justified */
CParser::TypeSpecifier CParser::checkTypeSpecifier(TypeSpecifier ts,
	TypeSpecifier add, _ANTLRTokenPtr tok) {
	// the first one is always valid
	if (ts==tsINVALID) return add;
	// nothing may be used twice
	if (ts&add) goto err;
	// void/float/union/struct/enum/typedef-name only alone
	if (ts&(tsVOID|tsFLOAT|tsUNION|tsSTRUCT|tsENUM|tsTYPEID) ||
		add&(tsVOID|tsFLOAT|tsUNION|tsSTRUCT|tsENUM|tsTYPEID)) goto err;
	// at most one of short/long
	if ((ts&(tsSHORT|tsLONG)) && (add&(tsSHORT|tsLONG))) goto err;
	// long may be used with double
	if (ts&tsDOUBLE && add!=tsLONG) goto err;
	if (add&tsDOUBLE && ts!=tsLONG) goto err;
	// at most one of signed/unsigned
	if (ts&(tsSIGNED|tsUNSIGNED) && add&(tsSIGNED|tsUNSIGNED)) goto err;
	// char can be neither int nor short nor long
	if ((ts&tsCHAR && add&(tsINT|tsSHORT|tsLONG)) ||
		(ts&(tsINT|tsSHORT|tsLONG) && add&tsCHAR)) goto err;
	return TypeSpecifier(ts|add);
err:
	synError("multiple incompatible type specifiers", tok);
	return ts;
}
